// ****************** MUST INCLUDE LINES BELOW*****************************
#include <lib6302.h>
// Init library and set pwm freq, DAC resolution, ADC resolution, and teensy 3.6-3.2 bit rate.
// PWM frequency affects DAC resolution, higher frequency, lower resolution.
// For 8-bit PWM: Teensy 3.6-234375, Teensy 3.2-187500!!!1
// Divide by 2,4,8 or 16 for 9,10,11, or 12 bits.
// WARNING 12 bit pwm is in audio range for young listeners.
#define MYSERIAL Serial
lib6302 my6302(234375/4,12,16,&Serial);
// Set up the analog-to-digital library, with a buffersize
#define sampleBufN 500
LIB6302_ADC(sampleBufN+100);  // ADC sample buffer of 100+sampleBufN
// ****************** MUST INCLUDE LINES ABOVE*******1**********************

// Constants for Users to Modify 
// ************************************************************
#define MODE 0  // Mode 0:plt meas+cmd, Mode 1:plt meas+desired, Mode 2:coil testing!
#define highD 0.0 // Toggle high (0.01->0.05)
#define lowD -0.0 // Toggle low  (-0.01=>-0.05)
#define sensorOffset 0.6 // Set so umbrella "hovers" at "d=0'.

// Might need to change some of these
#define KPNOM 0  // Nominal Kp value, added to value sent from plotter
#define KDNOM 0  // Nominal Kd value, added to value sent from plotter
#define mBack 1
#define controllerUpdatePeriod 50 // Update period in microseconds

#if MODE == 2  // Test coil mode 
#define togglePeriod 12.5e-3 // 12500 usecs between lowD-highD toggle
#define plotIncrement 1  // 50usecs between plot samples.
#else
#define togglePeriod 1.0 // 1 second between lowD-highD toggle
#define plotIncrement 40 // 2 milliseconds between plot samples.
#endif


// ************************************************************************
// End of Constants and code for Users to Modify 
// SEE END OF FILE FOR USER-ALTERABLE CONTROL LOOP!!

// *******SETTING UP THE MONITOR DISPLAY (SLIDERS AND PLOTS). *************
// Sliders
String sliders = "S~Pause~0~1~S~Kp~0~10~S~Kd~0~5";
#define PA 0
#define KP 1
#define KD 2

// Plots: US~E SPACES BETWEEN LABELS FOR MULTIPLOTS!!
String plotMag = "P~MagField CMD~-1.0~1.0~500~2~";
String plotDs = "P~Dmeas Ddes~-1~1~500~2~";
String plotDandCmd = "P~Dmeas~-1.0~1.0~500~1~P~Cmd~-1~1~500~1~";
#define PLOT_M 0

#if MODE == 2
String monPlots = plotMag;
#define PLOT_D 2
#define PLOT_C 1
#elif MODE == 1
String monPlots = plotDs;
#define PLOT_D 1
#define PLOT_C 2
#else
String monPlots = plotDandCmd;
#define PLOT_D 2
#define PLOT_C 1
#endif

// Configuration string with surrounding limits.
String config_message = sliders + monPlots;

#define zeroParamThresh 0.005
#define scaleParam 1.005
float analog2param(int pinA) {
  float retval = my6302.adcGetAvg(pinA) - zeroParamThresh;
  retval = scaleParam * max(retval, 0.0);
  return(retval);
}

// Control loop setup
// ************************************************************************

elapsedMicros loopTimer = 0;  // Loop runs at least 10xsecond
float mDeltaT = 0.0;
float deltaT = 0.0;
float headRoom;
float nomCmd = 0.0;
uint32_t loopCntr = 0;
boolean even = false;

void setup() {
  MYSERIAL.begin(1000000);
  my6302.setup(config_message, togglePeriod);
  my6302.startPWM(true); // start parallel drive pwm, A+B in sync.
  my6302.setupADC(4,true);
  nomCmd = 20.0*(2.0*analog2param(A3) - 1.0);
  loopTimer = 0;
  loopCntr = 0;
  deltaT = float(controllerUpdatePeriod)*1.0e-6;
  mDeltaT = mBack*deltaT;
  headRoom = controllerUpdatePeriod;
}

//********************** Main USER ALTERABLE Control Loop **********************
//**************************************************************************
int loopCounter = 0;
float pastError[mBack+1];  // Stores past errors for deriv approx
float pastCmd[mBack+1];
float fineCmd = 0;
float coarseCmd = 0;

void loop() {  

  // Initialize: check for inputs, flip toggle, etc.
  my6302.startloop();

  // Get parameters.
  float pauseVal= my6302.getSlider(PA);
  float Kp = KPNOM + my6302.getSlider(KP);
  float Kd = KDNOM + my6302.getSlider(KD);
  
  // Headroom is faction of period left until next sensor read.
  float headRoomThis = float(controllerUpdatePeriod - int(loopTimer));
  headRoom = headRoomThis;
  
  // Wait exactly one deltaT since last sensor read.
  while (int(loopTimer) < controllerUpdatePeriod) {};
  loopTimer = 0;
  
  // Read the optical distance or mag field sensor
  float meas = 0.0;
  if(MODE == 2) {  // Measure coil field
    meas = 4.0*(2.0*my6302.adcGetAvg(A1) - 1.0);
  }
  else {  // Measure distance sensor
    meas = 4*my6302.adcGetAvg(A0) - sensorOffset;
  }
    
  // Toggle desired value.
  float desired = highD;
  float flip = my6302.getFlip();
  if(flip > 0) desired = lowD;

  // Compute the errors
  float error = (desired - meas);
  
  // Save backward angles for delta/deltaT calculation.
  for (int i = mBack; i > 0; i--) {
    pastError[i] = pastError[i-1];
    pastCmd[i] = pastCmd[i-1];
  }
  pastError[0] = error;

  // Compute (error - oldError)/deltaT.
  float delDt = (error -pastError[mBack])/mDeltaT;
  
  // Compute feedback contribution to motor cmd = Kp*Err+Kd*D(Err)/Dt
  float dCmd = Kp*error + Kd*delDt; // FIX THIS FOR LEAD!
  pastCmd[0] = dCmd;

  if(MODE == 2) { // Test mode, then toggle between maximums.
    if(flip > 0) dCmd = 10.0 + nomCmd;
    else dCmd = -10.0 + nomCmd;
  }

  // Write motor command Hbridge, return clipped value.
  float cmdCL = my6302.hbridgeBipolar(nomCmd - dCmd, pwmB1_T,pwmB2_T); // B1_T->B2_T flips motor dir

  // Status and nominal updates every plotincrement cycle.
  if(loopCounter <  plotIncrement) loopCounter++;
  else {
    loopCounter = 1;
    // Read in potentiometers to set nominal cmd.
    if(even) {
      coarseCmd = 20.0*(2.0*analog2param(A3) - 1.0);
      even = false;
    } else {
      fineCmd = 1.0*(2.0*analog2param(A2) - 1.0);
      even = true;
    }
    nomCmd = coarseCmd+fineCmd;

    // If not paused, send status
    if(pauseVal == 0) {
      // Send data to monitor.
      float loopStatus[3];
      loopStatus[PLOT_M] = meas; //(headRoom < 0) ?  headRoom: meas;
      loopStatus[PLOT_D] = desired;
      loopStatus[PLOT_C] = cmdCL;
      my6302.sendStatus(2,loopStatus); 
    }
  }
}




  


  
